"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("./Type");
var Support_1 = require("./Support");
var TypeBuilder_1 = require("./TypeBuilder");
var Graph_1 = require("./Graph");
var TypeAttributeStore = /** @class */ (function () {
    function TypeAttributeStore(_typeGraph) {
        this._typeGraph = _typeGraph;
        this._attributeKinds = immutable_1.Set();
        this._values = [];
        this._topLevelValues = immutable_1.Map();
    }
    TypeAttributeStore.prototype.registerAttributeKind = function (name) {
        Support_1.assert(!this._attributeKinds.has(name), "Cannot register a type attribute more than once");
        this._attributeKinds = this._attributeKinds.add(name);
    };
    TypeAttributeStore.prototype.getTypeIndex = function (t) {
        var tref = t.typeRef;
        Support_1.assert(tref.graph === this._typeGraph, "Using the wrong type attribute store");
        return tref.getIndex();
    };
    TypeAttributeStore.prototype.setInMap = function (maybeMap, name, value) {
        if (!this._attributeKinds.has(name)) {
            return Support_1.panic("Unknown attribute " + name);
        }
        if (maybeMap === undefined) {
            maybeMap = immutable_1.Map();
        }
        maybeMap = maybeMap.set(name, value);
        return maybeMap;
    };
    TypeAttributeStore.prototype.set = function (name, t, value) {
        var index = this.getTypeIndex(t);
        while (index >= this._values.length) {
            this._values.push(undefined);
        }
        this._values[index] = this.setInMap(this._values[index], name, value);
    };
    TypeAttributeStore.prototype.setForTopLevel = function (attributeName, topLevelName, value) {
        var maybeMap = this._topLevelValues.get(topLevelName);
        this._topLevelValues = this._topLevelValues.set(topLevelName, this.setInMap(maybeMap, attributeName, value));
    };
    TypeAttributeStore.prototype.tryGetInMap = function (maybeMap, name) {
        Support_1.assert(this._attributeKinds.has(name), "Attribute " + name + " not defined");
        if (maybeMap === undefined) {
            return undefined;
        }
        return maybeMap.get(name);
    };
    TypeAttributeStore.prototype.tryGet = function (name, t) {
        var index = this.getTypeIndex(t);
        return this.tryGetInMap(this._values[index], name);
    };
    TypeAttributeStore.prototype.tryGetForTopLevel = function (attributeName, topLevelName) {
        return this.tryGetInMap(this._topLevelValues.get(topLevelName), attributeName);
    };
    return TypeAttributeStore;
}());
exports.TypeAttributeStore = TypeAttributeStore;
var TypeAttributeStoreView = /** @class */ (function () {
    function TypeAttributeStoreView(_attributeStore, _attributeName) {
        this._attributeStore = _attributeStore;
        this._attributeName = _attributeName;
        _attributeStore.registerAttributeKind(_attributeName);
    }
    TypeAttributeStoreView.prototype.set = function (t, value) {
        this._attributeStore.set(this._attributeName, t, value);
    };
    TypeAttributeStoreView.prototype.setForTopLevel = function (name, value) {
        this._attributeStore.setForTopLevel(this._attributeName, name, value);
    };
    TypeAttributeStoreView.prototype.tryGet = function (t) {
        return this._attributeStore.tryGet(this._attributeName, t);
    };
    TypeAttributeStoreView.prototype.get = function (t) {
        return Support_1.defined(this.tryGet(t));
    };
    TypeAttributeStoreView.prototype.tryGetForTopLevel = function (name) {
        return this._attributeStore.tryGetForTopLevel(this._attributeName, name);
    };
    TypeAttributeStoreView.prototype.getForTopLevel = function (name) {
        return Support_1.defined(this.tryGetForTopLevel(name));
    };
    return TypeAttributeStoreView;
}());
exports.TypeAttributeStoreView = TypeAttributeStoreView;
var TypeGraph = /** @class */ (function () {
    function TypeGraph(typeBuilder) {
        var _this = this;
        // FIXME: OrderedMap?  We lose the order in PureScript right now, though,
        // and maybe even earlier in the TypeScript driver.
        this._topLevels = immutable_1.Map();
        this.freeze = function (topLevels, types, typeNames) {
            Support_1.assert(!_this.isFrozen, "Tried to freeze TypeGraph a second time");
            Support_1.assert(types.every(function (t) { return t.typeRef.graph === _this; }), "Trying to freeze a graph with types that don't belong in it");
            // The order of these three statements matters.  If we set _typeBuilder
            // to undefined before we deref the TypeRefs, then we need to set _types
            // before, also, because the deref will call into typeAtIndex, which requires
            // either a _typeBuilder or a _types.
            _this._types = types;
            _this._typeBuilder = undefined;
            _this._topLevels = topLevels.map(function (tref) { return tref.deref()[0]; });
            for (var i = 0; i < types.length; i++) {
                var maybeNames = typeNames[i];
                if (maybeNames === undefined)
                    continue;
                _this._namesStoreView.set(types[i], maybeNames);
            }
        };
        this.allNamedTypes = function (childrenOfType) {
            return _this.filterTypes(Type_1.isNamedType, childrenOfType, true);
        };
        this.allNamedTypesSeparated = function (childrenOfType) {
            var types = _this.allNamedTypes(childrenOfType);
            return Type_1.separateNamedTypes(types);
        };
        this.allTypesUnordered = function () {
            Support_1.assert(_this.isFrozen, "Tried to get all graph types before it was frozen");
            return immutable_1.Set(Support_1.defined(_this._types));
        };
        this._typeBuilder = typeBuilder;
        this.attributeStore = new TypeAttributeStore(this);
        this._namesStoreView = new TypeAttributeStoreView(this.attributeStore, "names");
    }
    Object.defineProperty(TypeGraph.prototype, "isFrozen", {
        get: function () {
            return this._typeBuilder === undefined;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TypeGraph.prototype, "topLevels", {
        get: function () {
            Support_1.assert(this.isFrozen, "Cannot get top-levels from a non-frozen graph");
            return Support_1.defined(this._topLevels);
        },
        enumerable: true,
        configurable: true
    });
    TypeGraph.prototype.atIndex = function (index) {
        if (this._typeBuilder !== undefined) {
            return this._typeBuilder.atIndex(index);
        }
        var t = Support_1.defined(this._types)[index];
        return [t, this._namesStoreView.tryGet(t)];
    };
    TypeGraph.prototype.filterTypes = function (predicate, childrenOfType, topDown) {
        var seen = immutable_1.Set();
        var types = immutable_1.List();
        function addFromType(t) {
            if (seen.has(t))
                return;
            seen = seen.add(t);
            var required = predicate === undefined || predicate(t);
            if (topDown && required) {
                types = types.push(t);
            }
            var children = childrenOfType !== undefined ? childrenOfType(t) : t.children;
            children.forEach(addFromType);
            if (!topDown && required) {
                types = types.push(t);
            }
        }
        this.topLevels.forEach(addFromType);
        return types.toOrderedSet();
    };
    // Each array in `replacementGroups` is a bunch of types to be replaced by a
    // single new type.  `replacer` is a function that takes a group and a
    // TypeBuilder, and builds a new type with that builder that replaces the group.
    // That particular TypeBuilder will have to take as inputs types in the old
    // graph, but return types in the new graph.  Recursive types must be handled
    // carefully.
    TypeGraph.prototype.rewrite = function (stringTypeMapping, replacementGroups, replacer) {
        if (replacementGroups.length === 0)
            return this;
        return new TypeBuilder_1.GraphRewriteBuilder(this, stringTypeMapping, replacementGroups, replacer).finish();
    };
    TypeGraph.prototype.garbageCollect = function () {
        return new TypeBuilder_1.GraphRewriteBuilder(this, TypeBuilder_1.NoStringTypeMapping, [], function (_t, _b) {
            return Support_1.panic("This shouldn't be called");
        }).finish();
    };
    TypeGraph.prototype.makeGraph = function (invertDirection, childrenOfType) {
        return new Graph_1.Graph(Support_1.defined(this._types), invertDirection, childrenOfType);
    };
    return TypeGraph;
}());
exports.TypeGraph = TypeGraph;
function noneToAny(graph, stringTypeMapping) {
    var noneTypes = graph.allTypesUnordered().filter(function (t) { return t.kind === "none"; });
    if (noneTypes.size === 0) {
        return graph;
    }
    Support_1.assert(noneTypes.size === 1, "Cannot have more than one none type");
    return graph.rewrite(stringTypeMapping, [noneTypes.toArray()], function (_, builder, forwardingRef) {
        return builder.getPrimitiveType("any", forwardingRef);
    });
}
exports.noneToAny = noneToAny;
